package main

import (
	"errors"
	"fmt"
	"html/template"
	"net/http"
	"sync"

	log "github.com/sirupsen/logrus"

	"git.c3pb.de/gbe/invinoveritas/auth"
	"git.c3pb.de/gbe/invinoveritas/session"
)

func (h Handler) userCreate(w http.ResponseWriter, r *http.Request) (string, error) {
	name := r.PostForm.Get("new-user")
	if name == "" {
		return "", errors.New("name can't be empty")
	}

	if len(name) > 80 {
		return "", fmt.Errorf("user name too long %q", name)
	}

	generatedPassword, err := h.ap.CreateUser(r.Context(), name)
	if err != nil {
		return "", err
	}

	return generatedPassword, nil
}

func (h Handler) user() http.Handler {
	var (
		once sync.Once
		tpl  *template.Template
	)

	type templateData struct {
		User        auth.User
		AllUsers    []string // User names
		AllSessions []session.Info
		FormErrors  map[string]error
		Success     bool
		ShowResult  bool

		GeneratedPassword string // For new users
		CreatedUser       string // Name for new users
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		once.Do(func() {
			tpl = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/user.tpl"))
		})

		user := auth.Get(r)

		data := templateData{
			User:       user,
			FormErrors: make(map[string]error),
		}

		if r.Method == "GET" {
			var err error

			data.AllUsers, err = h.ap.ListUsers(r.Context())
			if err != nil {
				httpError(w, "can't get list of all users", err, http.StatusInternalServerError)
				return
			}

			data.AllSessions, err = h.ap.ListSessions(r.Context())
			if err != nil {
				httpError(w, "can't get list of sessions", err, http.StatusInternalServerError)
				return
			}

			err = tpl.ExecuteTemplate(w, "user.tpl", data)
			if err != nil {
				log.WithError(err).
					Error("can't execute details template")
			}

			return
		}

		err := r.ParseForm()
		if err != nil {
			httpError(w, "can't parse POST form", err, http.StatusInternalServerError)
			return
		}

		if r.Method == "DELETE" {
			err := h.ap.DeleteUser(r.Context(), r.PostForm.Get("username"))
			if err != nil {
				data.FormErrors["DeleteUser"] = err
			}

			return
		}

		defer func() {
			if err != nil {
				// We're in error handling mode, error has already been reported
				return
			}

			data.AllUsers, err = h.ap.ListUsers(r.Context())
			if err != nil {
				httpError(w, "can't get list of all users", nil, http.StatusInternalServerError)
				return
			}

			err = tpl.ExecuteTemplate(w, "user.tpl", data)
			if err != nil {
				log.WithError(err).
					Error("can't execute details template")
			}
		}()

		if r.Method != "POST" {
			err = errors.New(r.Method)
			httpError(w, "bad method", err, http.StatusMethodNotAllowed)
			return
		}

		switch r.PostForm.Get("action") {
		case "update":
			data.ShowResult = true
		case "create":
			generatedPw, err := h.userCreate(w, r)
			if err != nil {
				data.FormErrors["NewUser"] = err
			}

			data.GeneratedPassword = generatedPw
			data.CreatedUser = r.PostForm.Get("new-user")

			log.WithField("name", data.CreatedUser).
				Info("user created")

			return
		case "logout":
			cookie, err := r.Cookie("session")
			if err != nil {
				// We shouldn't end up here: if we do, someone managed to bypass the authentication
				// middleware.
				log.WithError(err).
					Error("can't get session token: possible auth bypass")

				http.Error(w, err.Error(), http.StatusInternalServerError)

				return
			}

			err = h.ap.DeleteSession(r.Context(), cookie.Value)
			if err != nil {
				log.WithError(err).
					Error("can't delete session")

				http.Error(w, err.Error(), http.StatusInternalServerError)
				return
			}

			http.Redirect(w, r, "/", http.StatusFound)

			return
		default:
			httpError(w, "unknown action", nil, http.StatusBadRequest)
			return
		}

		if r.PostForm.Get("new-pw") != r.PostForm.Get("new-pw-repeat") {
			data.FormErrors["NewPW"] = errors.New("new passwords don't match")
			return
		}

		if len(r.PostForm.Get("new-pw")) == 0 {
			data.FormErrors["NewPW"] = errors.New("new password can't be empty")
			return
		}

		err = h.ap.UpdatePassword(r.Context(), user.Name, r.PostForm.Get("old-pw"), r.PostForm.Get("new-pw"))
		if err != nil {
			data.FormErrors["OldPW"] = err
			return
		}

		data.Success = true
	})
}
